-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;		
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity gfunc_modified_ppl4 is
	generic ( 	iw : integer := 32; 
				h : integer := 256);
	port(
		clk : in std_logic;
		ain : in std_logic_vector(iw-1 downto 0);
		bin : in  std_logic_vector(iw-1 downto 0);
      	cin : in  std_logic_vector(iw-1 downto 0);
      	din : in  std_logic_vector(iw-1 downto 0);      
		const_0 : in  std_logic_vector(iw-1 downto 0);	
		const_1 : in  std_logic_vector(iw-1 downto 0);
      	aout : out std_logic_vector(iw-1 downto 0);
      	bout : out std_logic_vector(iw-1 downto 0);
      	cout : out std_logic_vector(iw-1 downto 0);
      	dout : out std_logic_vector(iw-1 downto 0)
		);		 
end gfunc_modified_ppl4;

architecture struct of gfunc_modified_ppl4 is	
	type pipelined_reg_type is array ( 1 to 3 ) of std_logic_vector(iw-1 downto 0);
	signal a_reg, b_reg, c_reg, d_reg : pipelined_reg_type;
	
	-- stage 0
	signal a_add1, d_xor1 : std_logic_vector(iw-1 downto 0);
	
	-- stage 1
	signal d_rot1, c_add1, b_xor1 : std_logic_vector(iw-1 downto 0);
	
	-- stage 2
	signal b_rot1, a_add2, d_xor2 : std_logic_vector(iw-1 downto 0);
	
	-- stage 3
	signal d_rot2, c_add2, b_xor2, b_rot2 : std_logic_vector(iw-1 downto 0);
begin 
	
	a_add1 <= const_0 + ain + bin;
	d_xor1 <= a_add1 xor din;
	
	stage1 : process ( clk ) 
	begin
		if rising_edge( clk ) then
			a_reg(1) <= a_add1;
			b_reg(1) <= bin;
			c_reg(1) <= cin;
			d_reg(1) <= d_xor1;
		end if;
	end process;
	
	c_add1 <= d_rot1 + c_reg(1); 
	b_xor1 <= c_add1 xor b_reg(1);
	
	stage2 : process ( clk ) 
	begin
		if rising_edge( clk ) then
			a_reg(2) <= a_reg(1);
			b_reg(2) <= b_xor1;
			c_reg(2) <= c_add1;
			d_reg(2) <= d_rot1;
		end if;
	end process;
	
	a_add2 <= b_rot1 + a_reg(2) + const_1;
	d_xor2 <= a_add2 xor d_reg(2);
	
	stage3 : process ( clk ) 
	begin
		if rising_edge( clk ) then
			a_reg(3) <= a_add2;
			b_reg(3) <= b_rot1;
			c_reg(3) <= c_reg(2);
			d_reg(3) <= d_xor2;
		end if;
	end process;		 
		
	c_add2 <= d_rot2 + c_reg(3);
	b_xor2 <= c_add2 xor b_reg(3);		
	
	arch_32 : if h <= 256 generate
		d_rot1 <= rolx(d_reg(1), 32-g_rot_arch32(0));
		b_rot1 <= rolx(b_reg(2), 32-g_rot_arch32(1));
		d_rot2 <= rolx(d_reg(3), 32-g_rot_arch32(2));
		b_rot2 <= rolx(b_xor2, 32-g_rot_arch32(3));
	end generate;
	
	arch_64 : if h > 256 generate
		d_rot1 <= rolx(d_reg(1), 64-g_rot_arch64(0));
		b_rot1 <= rolx(b_reg(2), 64-g_rot_arch64(1));
		d_rot2 <= rolx(d_reg(3), 64-g_rot_arch64(2));
		b_rot2 <= rolx(b_xor2, 64-g_rot_arch64(3));
	end generate;	
	
	
	-- OUTPUT
	aout <= a_reg(3);
	bout <= b_rot2;
	cout <= c_add2;
	dout <= d_rot2;
end struct;

----===================================
--- PPL 2
----===================================
         --

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;		
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity gfunc_modified_ppl2 is
	generic ( 	iw : integer := 32; 
				h : integer := 256);
	port(
		clk : in std_logic;
		ain : in std_logic_vector(iw-1 downto 0);
		bin : in  std_logic_vector(iw-1 downto 0);
      	cin : in  std_logic_vector(iw-1 downto 0);
      	din : in  std_logic_vector(iw-1 downto 0);      
		const_0 : in  std_logic_vector(iw-1 downto 0);	
		const_1 : in  std_logic_vector(iw-1 downto 0);
      	aout : out std_logic_vector(iw-1 downto 0);
      	bout : out std_logic_vector(iw-1 downto 0);
      	cout : out std_logic_vector(iw-1 downto 0);
      	dout : out std_logic_vector(iw-1 downto 0)
		);		 
end gfunc_modified_ppl2;

architecture struct of gfunc_modified_ppl2 is	
	signal a_reg, b_reg, c_reg, d_reg : std_logic_vector(iw-1 downto 0);	
	-- stage 0
	signal a_add1, d_xor1 : std_logic_vector(iw-1 downto 0);	
	-- stage 1
	signal d_rot1, c_add1, b_xor1 : std_logic_vector(iw-1 downto 0);	
	-- stage 2
	signal b_rot1, a_add2, d_xor2 : std_logic_vector(iw-1 downto 0);	
	-- stage 3
	signal d_rot2, c_add2, b_xor2, b_rot2 : std_logic_vector(iw-1 downto 0);
begin 	
	a_add1 <= const_0 + ain + bin;
	d_xor1 <= a_add1 xor din;	
	c_add1 <= d_rot1 + cin; 
	b_xor1 <= c_add1 xor bin;
	a_add2 <= b_rot1 + a_add1 + const_1;
	
	
	pipelined_reg : process ( clk ) 
	begin
		if rising_edge( clk ) then
			a_reg <= a_add2;
			b_reg <= b_rot1;
			c_reg <= c_add1;
			d_reg <= d_rot1;
		end if;
	end process;		 
	
	d_xor2 <= a_reg xor d_reg;	
	c_add2 <= d_rot2 + c_reg;
	b_xor2 <= c_add2 xor b_reg;		
	
	arch_32 : if h <= 256 generate
		d_rot1 <= rolx(d_xor1, 32-g_rot_arch32(0));
		b_rot1 <= rolx(b_xor1, 32-g_rot_arch32(1));
		d_rot2 <= rolx(d_xor2, 32-g_rot_arch32(2));
		b_rot2 <= rolx(b_xor2, 32-g_rot_arch32(3));
	end generate;
	
	arch_64 : if h > 256 generate
		d_rot1 <= rolx(d_xor1, 64-g_rot_arch64(0));
		b_rot1 <= rolx(b_xor1, 64-g_rot_arch64(1));
		d_rot2 <= rolx(d_xor2, 64-g_rot_arch64(2));
		b_rot2 <= rolx(b_xor2, 64-g_rot_arch64(3));
	end generate;	
	
	
	-- OUTPUT
	aout <= a_reg;
	bout <= b_rot2;
	cout <= c_add2;
	dout <= d_rot2;
end struct;




----===================================
--- GFUNCx8 PPL 2
----===================================
-- Developer note : 
--	The main difference from GFUNCx4 PPL2
-- 	is the location of pipeline registers.
-- 	In GFUNCx4, the pipeline registers locate after
--	the second addition of constant and message to A.
--	In GFUNCx8, the pipeline registers locate in the middle
--	of the GFUNC. This is after the first B rotation

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;		
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity gfunc8x_modified_ppl2 is
	generic ( 	iw : integer := 32; 
				h : integer := 256);
	port(
		clk : in std_logic;
		ain : in std_logic_vector(iw-1 downto 0);
		bin : in  std_logic_vector(iw-1 downto 0);
      	cin : in  std_logic_vector(iw-1 downto 0);
      	din : in  std_logic_vector(iw-1 downto 0);      
		const_0 : in  std_logic_vector(iw-1 downto 0);	
		const_1 : in  std_logic_vector(iw-1 downto 0);
      	aout : out std_logic_vector(iw-1 downto 0);
      	bout : out std_logic_vector(iw-1 downto 0);
      	cout : out std_logic_vector(iw-1 downto 0);
      	dout : out std_logic_vector(iw-1 downto 0)
		);		 
end gfunc8x_modified_ppl2;

architecture struct of gfunc8x_modified_ppl2 is	
	signal a_reg, b_reg, c_reg, d_reg : std_logic_vector(iw-1 downto 0);	
	-- stage 0
	signal a_add1, d_xor1 : std_logic_vector(iw-1 downto 0);	
	-- stage 1
	signal d_rot1, c_add1, b_xor1 : std_logic_vector(iw-1 downto 0);	
	-- stage 2
	signal b_rot1, a_add2, d_xor2 : std_logic_vector(iw-1 downto 0);	
	-- stage 3
	signal d_rot2, c_add2, b_xor2, b_rot2 : std_logic_vector(iw-1 downto 0);
begin 	
	a_add1 <= const_0 + ain + bin;
	d_xor1 <= a_add1 xor din;	
	c_add1 <= d_rot1 + cin; 
	b_xor1 <= c_add1 xor bin;
	
	
	
	pipelined_reg : process ( clk ) 
	begin
		if rising_edge( clk ) then
			a_reg <= a_add1;
			b_reg <= b_rot1;
			c_reg <= c_add1;
			d_reg <= d_rot1;
		end if;
	end process;		 
	
	a_add2 <= b_reg + a_reg + const_1;	
	d_xor2 <= a_add2 xor d_reg;	
	c_add2 <= d_rot2 + c_reg;
	b_xor2 <= c_add2 xor b_reg;		
	
	arch_32 : if h <= 256 generate
		d_rot1 <= rolx(d_xor1, 32-g_rot_arch32(0));
		b_rot1 <= rolx(b_xor1, 32-g_rot_arch32(1));
		d_rot2 <= rolx(d_xor2, 32-g_rot_arch32(2));
		b_rot2 <= rolx(b_xor2, 32-g_rot_arch32(3));
	end generate;
	
	arch_64 : if h > 256 generate
		d_rot1 <= rolx(d_xor1, 64-g_rot_arch64(0));
		b_rot1 <= rolx(b_xor1, 64-g_rot_arch64(1));
		d_rot2 <= rolx(d_xor2, 64-g_rot_arch64(2));
		b_rot2 <= rolx(b_xor2, 64-g_rot_arch64(3));
	end generate;	
	
	
	-- OUTPUT
	aout <= a_add2;
	bout <= b_rot2;
	cout <= c_add2;
	dout <= d_rot2;
end struct;


------------------------------------------
------------------------------------------
------------------------------------------
------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;		
use work.sha3_pkg.all;
use work.sha3_blake_package.all;


entity gfunc_modified is
	generic ( 	iw : integer := 32; 
				h : integer := 256	);
	port(
		ain : in std_logic_vector(iw-1 downto 0);
		bin : in  std_logic_vector(iw-1 downto 0);
      	cin : in  std_logic_vector(iw-1 downto 0);
      	din : in  std_logic_vector(iw-1 downto 0);      
		const_0 : in  std_logic_vector(iw-1 downto 0);	
		const_1 : in  std_logic_vector(iw-1 downto 0);
      	aout : out std_logic_vector(iw-1 downto 0);
      	bout : out std_logic_vector(iw-1 downto 0);
      	cout : out std_logic_vector(iw-1 downto 0);
      	dout : out std_logic_vector(iw-1 downto 0)
		);		 
end gfunc_modified;

--}} end of automatically maintained section

architecture struct of gfunc_modified is
	signal csa_a1_1, csa_a1_2, csa_a2_1, csa_a2_2 : std_logic_vector(iw-1 downto 0);

	signal  D_hold2, D_hold3     :   STD_LOGIC_VECTOR(iw-1 downto 0);
	signal  hold_D1, hold_B2     :   STD_LOGIC_VECTOR(iw-1 downto 0);
	signal  hold_B1, hold_D2              :   STD_LOGIC_VECTOR(iw-1 downto 0);
	signal  D_hold5, D_hold6              :   STD_LOGIC_VECTOR(iw-1 downto 0);
	signal  hold_A1,  hold_C1             :   std_logic_vector (iw-1 downto 0); 
	signal  hold_A2, hold_C2	          :   std_logic_vector (iw-1 downto 0);
begin 
	
	D_hold2 <= din xor hold_A1;					 
	hold_C1 <= hold_D1+ cin;
	D_hold3 <= hold_C1 xor bin;
	
	D_hold5 <= hold_A2 xor hold_D1;
	hold_C2 <= hold_C1 + hold_D2;
	D_hold6 <= hold_C2 xor hold_B1;
	
	-- output data
	aout   <= hold_A2;
	bout   <= hold_B2;
	cout   <= hold_C2;
	dout   <= hold_D2;
	

	hold_A1 <= const_0 + ain + bin;
	hold_A2 <= const_1 + hold_A1 + hold_B1;
	
	
	arch_32 : if h <= 256 generate
		hold_d1 <= rolx(D_hold2, 32-g_rot_arch32(0));
		hold_B1 <= rolx(D_hold3, 32-g_rot_arch32(1));
		hold_D2 <= rolx(D_hold5, 32-g_rot_arch32(2));
		hold_B2 <= rolx(D_hold6, 32-g_rot_arch32(3));
	end generate;
	
	arch_64 : if h > 256 generate
		hold_d1 <= rolx(D_hold2, 64-g_rot_arch64(0));
		hold_B1 <= rolx(D_hold3, 64-g_rot_arch64(1));
		hold_D2 <= rolx(D_hold5, 64-g_rot_arch64(2));
		hold_B2 <= rolx(D_hold6, 64-g_rot_arch64(3));
	end generate;	
end struct;
